'''
Created on Oct 31, 2016

@author: nicolas
'''

ATTACHMENTS_DIR = "attachments"

LOG_LEVEL_DEBUG = "debug"
LOG_LEVEL_INFO = "info"
LOG_LEVEL_WARN = "warn"
LOG_LEVEL_ERROR = "error"

NEGATIVE_CHARS = "-^~"
